KEY__id = "id"
KEY__last_totp = "last_totp"
KEY__application = "application"
KEY__application_description = "application_description"
KEY__configuration = "configuration"
KEY__configuration_name = "name"
KEY__configuration_description = "configuration_description"
KEY__exception = "exception"
KEY__install_key = "install_key"
KEY__deletion_key = "deletion_key"
KEY__superjaaql_password = "superjaaql_password"
KEY__records_total = "records_total"
KEY__records_filtered = "records_filtered"
KEY__data = "data"
KEY__db_url = "db_url"
KEY__precedence = "precedence"
KEY__roles = "roles"
KEY__pre_auth_key = "pre_auth_key"
KEY__email = "email"
KEY__sort = "sort"
KEY__invite_key = "invite_key"
KEY__sign_up_data = "sign_up_data"
KEY__search = "search"
KEY__application_name = "name"
KEY__dataset = "dataset"
KEY__dataset_name = "name"
KEY__dataset_description = "dataset_description"
KEY__size = "size"
KEY__database = "database"
KEY__node_name = "name"
KEY__authorization = "authorization"
KEY__argument = "argument"
KEY__is_node = "is_node"
KEY__connection = "connection"
KEY__node = "node"
KEY__application_url = "url"
KEY__role = "role"
KEY__mfa_key = "mfa_key"
KEY__mfa_enabled = "mfa_enabled"
KEY__username = "username"
KEY__password = "password"
KEY__new_password = "new_password"
KEY__new_password_confirm = "new_password_confirm"
KEY__page = "page"
KEY__use_mfa = "use_mfa"
KEY__show_deleted = "show_deleted"
KEY__description = "description"
KEY__database_name = "name"
KEY__create = "create"
KEY__drop = "drop"
KEY__force_transactional = "force_transactional"
KEY__port = "port"
KEY__address = "address"
KEY__jaaql_name = "jaaql_name"
KEY__otp_uri = "otp_uri"
KEY__otp_qr = "otp_qr"
KEY__user_agent = "user_agent"
KEY__jaaql_otp_uri = "jaaql_otp_uri"
KEY__jaaql_otp_qr = "jaaql_otp_qr"
KEY__superjaaql_otp_uri = "superjaaql_otp_uri"
KEY__superjaaql_otp_qr = "superjaaql_otp_qr"

ATTR__deleted = "deleted"

HEADER__security = "Authentication-Token"

SQL__where = "WHERE"
SQL__order_by = "ORDER BY"
SQL__limit = "LIMIT"
SQL__offset = "OFFSET"
SQL__asc = "ASC"
SQL__desc = "DESC"
SQL__and = "AND"
SQL__or = "OR"
SQL__like = "LIKE"
SQL__eq = "="
SQL__paren_open = "("
SQL__paren_close = ")"
SQL__single_quote = "'"
SQL__double_quote = '"'
SQL__lower = "lower"
SQL__varchar_cast = "::varchar"
SQL__is_null = "is null"

SEPARATOR__comma_space = ", "
SEPARATOR__comma = ","
SEPARATOR__space = " "

JAAQL__arg_marker = ":"

ENCODING__utf = "UTF-8"
ENCODING__ascii = "ascii"
EXAMPLE__jwt = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.e30.yXvILkvUUCBqAFlAv6wQ1Q-QRAjfe3eSosO949U73Vo"

VERSION = "2.5.0"
